#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "gc.h"

#define HTX_MINSIZE 8
#define HTX_MAXSIZE (1U << 31)

#define JUMP(i, j) i += j++
#define JUMP_FIRST(i, j) j = 1, i += j++

static inline PREFIX(_hash_t) default_keyhash(PREFIX(_key_t) key)
{
	return (PREFIX(_hash_t))key;
}

static inline int default_keyeq(PREFIX(_key_t) a, PREFIX(_key_t) b)
{
	return a == b;
}

/* generic functions */

static inline PREFIX(_entry_t) *allocempty(unsigned int n)
{
	return GC_malloc(n * sizeof(PREFIX(_entry_t)));
}

static inline void entrycpy(PREFIX(_entry_t) *dst, const PREFIX(_entry_t) *src)
{
	dst->key = src->key;
	dst->value = src->value;
/*	memcpy(dst, src, sizeof(PREFIX(_entry_t)));*/
}

/* fill threshold = 3/4 */
#define HTX_LOG_THRES 2
static inline int checkfill(PREFIX(_t) *ht)
{
	if (ht->fill > ht->mask - (ht->mask >> HTX_LOG_THRES)) {
		PREFIX(_resize)(ht, ht->fill << 1);
		return 1;
	}
	return 0;
}

PREFIX(_t) *PREFIX(_alloc)(PREFIX(_hash_t) (*keyhash)(PREFIX(_key_t)), int (*keyeq)(PREFIX(_key_t), PREFIX(_key_t)))
{
	PREFIX(_t) *ht = GC_malloc(sizeof(PREFIX(_t)));
	
	return PREFIX(_init)(ht, keyhash, keyeq);
}

PREFIX(_t) *PREFIX(_init)(PREFIX(_t) *ht, PREFIX(_hash_t) (*keyhash)(PREFIX(_key_t)), int (*keyeq)(PREFIX(_key_t), PREFIX(_key_t)))
{
	memset(ht, 0, sizeof(PREFIX(_t)));
	ht->mask = HTX_MINSIZE - 1;
	ht->table = allocempty(ht->mask + 1);
	ht->keyhash = keyhash ? keyhash : default_keyhash;
	ht->keyeq = keyeq ? keyeq : default_keyeq;
	return ht;
}

PREFIX(_t) *PREFIX(_clear)(PREFIX(_t) *ht)
{
	PREFIX(_uninit)(ht);
	return PREFIX(_init)(ht, ht->keyhash, ht->keyeq);
}

PREFIX(_t) *PREFIX(_uninit)(PREFIX(_t) *ht)
{
	//free(ht->table);
	ht->table = NULL;
	return ht;
}

void PREFIX(_free)(PREFIX(_t) *ht)
{
	PREFIX(_uninit)(ht);
	//free(ht);
}

/* one lookup function to rule them all */
static PREFIX(_entry_t) *PREFIX(_lookup)(PREFIX(_t) *ht, PREFIX(_key_t) key, PREFIX(_hash_t) hash)
{
	unsigned int mask = ht->mask;
	unsigned int i = hash;
	unsigned int j;
	PREFIX(_entry_t) *table = ht->table;
	PREFIX(_entry_t) *entry = table + (i & mask);
#ifdef HTX_STAT
	++ht->lookup;
#endif
	assert(ht->table && key);
	if (PREFIX(_isempty)(entry) || ht->keyeq(entry->key, key))
		return entry;
	assert(ht->fill <= ht->mask);
	for (JUMP_FIRST(i, j); ; JUMP(i, j)) {
#ifdef HTX_STAT
		if (ht->keyhash(entry->key) == hash)
			++ht->samehash;
		++ht->jump;
#endif
		entry = table + (i & mask);
		if (PREFIX(_isempty)(entry) || ht->keyeq(entry->key, key))
			return entry;
	}
}

/* for copy and resize: entry->key is not in ht */
static void cleaninsert(PREFIX(_t) *ht, const PREFIX(_entry_t) *entry)
{
	unsigned int mask = ht->mask;
	unsigned int i = ht->keyhash(entry->key);
	unsigned int j;
	PREFIX(_entry_t) *table = ht->table;
	PREFIX(_entry_t) *newentry = table + (i & mask);
	
	if (PREFIX(_isused)(newentry))
		for (JUMP_FIRST(i, j); PREFIX(_isused)(newentry); JUMP(i, j))
			newentry = table + (i & mask);
	++ht->fill;
	entrycpy(newentry, entry);
}

PREFIX(_t) *PREFIX(_copy)(const PREFIX(_t) *ht)
{
	PREFIX(_t) *newht;
	PREFIX(_entry_t) *entry;
	unsigned int used;
	
	newht = calloc(1, sizeof(PREFIX(_t)));
	newht->mask = ht->mask;
	newht->table = allocempty(newht->mask + 1);
	for (entry = ht->table, used = ht->fill; used > 0; ++entry)
		if (PREFIX(_isused)(entry)) {
			--used;
			cleaninsert(newht, entry);
		}
	return newht;
}

PREFIX(_t) *PREFIX(_resize)(PREFIX(_t) *ht, unsigned int hint)
{
	unsigned int newsize;
	unsigned int oldused = ht->fill;
	PREFIX(_entry_t) *oldtable, *newtable, *entry;
	
	if (hint < oldused << 1)
		hint = oldused << 1;
	assert(hint <= HTX_MAXSIZE && hint > oldused);
	for (newsize = HTX_MINSIZE; newsize < hint; newsize <<= 1);
	newtable = allocempty(newsize);
#ifdef HTX_STAT
	++ht->resize;
#endif
	oldtable = ht->table;
	ht->table = newtable;
	ht->mask = newsize - 1;
	ht->fill = 0;
	for (entry = oldtable; oldused > 0; ++entry)
		if (PREFIX(_isused)(entry)) {
			--oldused;
			cleaninsert(ht, entry);
		}
	//free(oldtable);
	return ht;
}

int PREFIX(_contains)(PREFIX(_t) *ht, PREFIX(_key_t) key)
{
	PREFIX(_entry_t) *entry = PREFIX(_lookup)(ht, key, ht->keyhash(key));
	
	return PREFIX(_isused)(entry);
}

PREFIX(_value_t) PREFIX(_get)(PREFIX(_t) *ht, PREFIX(_key_t) key)
{
	PREFIX(_entry_t) *entry = PREFIX(_lookup)(ht, key, ht->keyhash(key));
	
	assert(PREFIX(_isused)(entry));
	return entry->value;
}

PREFIX(_entry_t) *PREFIX(_getentry)(PREFIX(_t) *ht, PREFIX(_key_t) key)
{
	PREFIX(_entry_t) *entry = PREFIX(_lookup)(ht, key, ht->keyhash(key));
	
	return PREFIX(_isused)(entry) ? entry : NULL;
}

PREFIX(_entry_t) *PREFIX(_insert)(PREFIX(_t) *ht, PREFIX(_key_t) key, PREFIX(_value_t) value)
{
	PREFIX(_entry_t) *entry = PREFIX(_lookup)(ht, key, ht->keyhash(key));
	
	if (PREFIX(_isused)(entry))
		return entry;
#ifdef HTX_STAT
	++ht->insert;
#endif
	++ht->fill;
	entry->key = key;
	entry->value = value;
	checkfill(ht);
	return NULL;
}

void PREFIX(_set)(PREFIX(_t) *ht, PREFIX(_key_t) key, PREFIX(_value_t) value)
{
	PREFIX(_entry_t) *entry = PREFIX(_insert)(ht, key, value);
	
	if (entry)
		entry->value = value;
}

#ifdef HTX_STAT
#include <stdio.h>
void PREFIX(_printstat)(const PREFIX(_t) *ht)
{
	printf("size: %u\nfill: %u\ninsert: %u\n\
resize: %u\njump: %u\nsamehash: %u\nlookup: %u\n",
	ht->mask + 1, ht->fill, ht->insert,
	ht->resize, ht->jump, ht->samehash, ht->lookup);
}
#endif
