#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "area.h"
#include "cmd.h"
#include "poly.h"

#ifdef WIN32
#include <malloc.h>
#endif

area_list_t **areamap = NULL;
int areamap_w = -1, areamap_h = -1;
double areamap_ox, areamap_oy, areamap_mx, areamap_my;

/* hash used for accelerating rendering:
 * area_list_t -> result area_list_t */
vptrht_t * area_render_hash = NULL;

void areamap_resize(int w, int h)
{
	areamap_w = w;
	areamap_h = h;

	if(areamap) free(areamap);
	areamap = malloc(w * h * sizeof *areamap);
	if(!areamap) abort();

	render_areas();
}

area_list_t * area_list_new(int count)
{
	area_list_t *p = malloc(count * sizeof(p->areas[0]) + sizeof *p);
	if(!p) abort();
	p->count = count;
	return p;
}

int area_list_index(area_list_t *p, area_t *area)
{
	int i;
	for(i=0; i<p->count; i++)
		if(p->areas[i] == area) return i;
	return -1;
}

area_list_t * area_list_copy(area_list_t *p)
{
	area_list_t *p2 = area_list_new(p->count);
	memcpy(p2->areas, p->areas, p->count * sizeof(p->areas[0]));
	return p2;
}

area_list_t * area_list_add(area_list_t *p, area_t *area)
{
	area_list_t *p2 = area_list_new(p->count + 1);

	int i, j;
	for(i=0, j=0; i<p->count;) {
		if(area && area < p->areas[i]) {
			p2->areas[j++] = area;
			area = NULL;
		} else {
			p2->areas[j++] = p->areas[i++];
		}
	}

	if(j == i) p2->areas[j] = area;

	return p2;
}

area_list_t * area_list_sub(area_list_t *p, area_t *area)
{
	assert(p->count > 0);
	area_list_t *p2 = area_list_new(p->count - 1);

	int i, j;
	for(i=0, j=0; i<p->count; i++) {
		if(p->areas[i] != area) p2->areas[j++] = p->areas[i];
	}
	assert(j == i-1);
	return p2;
}

int area_list_eq(area_list_t *p1, area_list_t *p2)
{
	if(p1->count != p2->count) return 0;
	return 0 == memcmp(p1->areas, p2->areas, p1->count * sizeof(p1->areas[0]));
}

static ptrhash_hash_t area_list_hash(area_list_t *p)
{
	return MurmurHash2(p, p->count * sizeof(p->areas[0]) + sizeof *p, 0);
}

ptrhash_t * area_list_hash_new()
{
	ptrhash_t *ht = ptrhash_alloc((ptrhash_hash_t (*) (ptrhash_key_t)) area_list_hash, (int (*) (ptrhash_key_t, ptrhash_key_t)) area_list_eq);
	return ht;
}

vptrht_t * area_ptr_hash_new()
{
	vptrht_t *ht = vptrht_alloc();
	return ht;
}

static ptrhash_t * area_list_master_hash = NULL;
static ptrhash_t * area_name_hash = NULL;

void init_area()
{
	area_list_master_hash = area_list_hash_new();
	area_name_hash = strhash_new();
	area_render_hash = area_ptr_hash_new();
}

/* retrieves list equivalent to parameter
 * from master hash
 * if it doesn't exist, list is copied and
 * put in hash */
area_list_t * area_list_store(area_list_t *l)
{
	ptrhash_entry_t *entry = ptrhash_getentry(area_list_master_hash, l);
	if(entry) return entry->value;

	l = area_list_copy(l);
	ptrhash_set(area_list_master_hash, l, l);

	return l;
}

void kill_areas()
{
	clear_area_list_store();

	ptrhash_entry_t *e;
	for(e = ptrhash_first(area_name_hash); e; e = ptrhash_next(area_name_hash, e)) {
		free(e->key);
		if(e->value) {
			area_t *a = e->value;
			free(a->p);
		}
		free(e->value);
	}
	ptrhash_clear(area_name_hash);

	render_areas();
}

void clear_area_list_store()
{
	ptrhash_entry_t *e;
	for(e = ptrhash_first(area_list_master_hash); e; e = ptrhash_next(area_list_master_hash, e)) {
		assert(e->key == e->value);
		free(e->key);
	}
	ptrhash_clear(area_list_master_hash);
}

area_list_t * area_list_store_add(area_list_t *p, area_t *area)
{
	area_list_t *nl = area_list_add(p, area);
	area_list_t *rl = area_list_store(nl);
	free(nl);
	return rl;
}

area_list_t * area_list_store_sub(area_list_t *p, area_t *area)
{
	area_list_t *nl = area_list_sub(p, area);
	area_list_t *rl = area_list_store(nl);
	free(nl);
	return rl;
}

void set_area(char *name, polygon_t *p)
{
	int i, j;
	for(i=0, j=0; i<p->npoints; i++, j+=2) {
		double nx, ny;
		transform(p->points[j], p->points[j+1], &nx, &ny);
		p->points[j] = nx;
		p->points[j+1] = ny;
	}

	ptrhash_entry_t *entry = ptrhash_getentry(area_name_hash, name);
	if(entry && entry->value) {
		free(name);
		area_t *olda = entry->value;
		area_render_unset(olda);
		free(olda->p);
		olda->p = p;
		area_render_set(olda);
	} else {
		if(entry) {
			free(name);
			name = entry->key;
		}
		area_t *a = malloc(sizeof *a);
		if(!a) abort();
		a->name = name;
		a->p = p;
		ptrhash_set(area_name_hash, name, a);
		area_render_set(a);
	}
}

int warp_area(char *name)
{
	ptrhash_entry_t *entry = ptrhash_getentry(area_name_hash, name);

	if(entry && entry->value) {
		area_t *a = entry->value;

		int checkpt(double x, double y)
		{
			area_list_t *al = areamap_query(areamap_ox + x * areamap_mx, areamap_oy + y * areamap_my);
			if(!al) return 0;
			int i;
			for(i=0; i<al->count; i++) {
				if(al->areas[i] == a) {
					return 1;
				}
			}
			return 0;
		}

		int i;
		double x = 0;
		double y = 0;
		int j = 0;
		for(i=0; i<a->p->npoints; i++) {
			x += a->p->points[j++];
			y += a->p->points[j++];
		}
		x /= a->p->npoints;
		y /= a->p->npoints;

		if(checkpt(x, y)) {
			SDL_WarpMouse((areamap_ox + x * areamap_mx)*2, (areamap_oy + y * areamap_my)*2);
			return 0;
		}

		j = 0;
		for(i=0; i<a->p->npoints; i++) {
			x = a->p->points[j++];
			y = a->p->points[j++];
			if(checkpt(x, y)) {
				SDL_WarpMouse((areamap_ox + x * areamap_mx)*2, (areamap_oy + y * areamap_my)*2);
				return 0;
			}
		}

		/* TODO: something more efficient perhaps? */
		int px, py;
		for(py = 0; py<areamap_h; py ++) {
			for(px = 0; px<areamap_w; px ++) {
				area_list_t *al = areamap_query(px, py);
				if(al) {
					for(i=0; i<al->count; i++) {
						if(al->areas[i] == a) {
							SDL_WarpMouse(px * 2, py * 2);
							return 0;
						}
					}
				}
			}
		}

		fprintf(stderr, "warp area error: area %s has no visible pixel\n", name);
		return 0;
	} else {
		fprintf(stderr, "warp area error: area %s not defined\n", name);
		return 1;
	}
}

void del_area(char *name)
{
	ptrhash_entry_t *entry = ptrhash_getentry(area_name_hash, name);
	if(entry && entry->value) {
		area_t *olda = entry->value;
		area_render_unset(olda);
		free(olda->p);
		free(olda);
		entry->value = NULL;
	}
}

void render_areas()
{
	memset(areamap, 0, areamap_w * areamap_h * sizeof *areamap);

	ptrhash_entry_t *e;
	for(e = ptrhash_first(area_name_hash); e; e = ptrhash_next(area_name_hash, e)) {
		if(e->value) area_render_set(e->value);
	}
}

area_list_t *areamap_query(int x, int y)
{
	if(!areamap) return NULL;
	if(x < 0 || y < 0 || x >= areamap_w || y >= areamap_h) return NULL;
	return areamap[y*areamap_w+x];
}

void area_render_set(area_t *a)
{
	vptrht_clear(area_render_hash);
	area_fillpoly(a, 1);
}

void area_render_unset(area_t *a)
{
	vptrht_clear(area_render_hash);
	area_fillpoly(a, 0);
}

static inline void area_hline(area_t *a, int set, area_list_t *id, int xmin, int xmax, int ypos)
{
	area_list_t **p = areamap + ypos * areamap_w + xmin;
	int i=xmin;

	area_list_t *np;

	if(set) {
		for(; i<xmax; i++, p++) {
			if(*p == NULL) {
				*p = id;
			} else {
				np = vptrht_get(area_render_hash, *p);
				if(np) {
					*p = np;
				} else {
					np = area_list_store_add(*p, a);
					vptrht_set(area_render_hash, *p, np);
					*p = np;
				}
			}
		}
	} else {
		for(; i<xmax; i++, p++) {
			assert((*p) != NULL);
			if(*p == id) {
				*p = NULL;
			} else {
				np = vptrht_get(area_render_hash, *p);
				if(np) {
					*p = np;
				} else {
					np = area_list_store_sub(*p, a);
					vptrht_set(area_render_hash, *p, np);
					*p = np;
				}
			}
		}
	}
}

void area_fillpoly(area_t *a, int set)
{
	int n = a->p->npoints;
	gfx_polypt pt = alloca(n * 2 * sizeof(int));
	area_list_t *l_id = alloca(sizeof(area_t*) + sizeof *l_id);

	l_id->count = 1;
	l_id->areas[0] = a;
	l_id = area_list_store(l_id);

	int i, j=0;
	for(i=0; i<n; i++) {
		pt[i][0] = areamap_ox + a->p->points[j++] * areamap_mx;
		pt[i][1] = areamap_oy + a->p->points[j++] * areamap_my;
	}

	void hline(int x1, int x2, int y)
	{
		area_hline(a, set, l_id, x1, x2, y);
	}

	scanpoly(0, areamap_w - 1, 0, areamap_h - 1, n, pt, hline);
}
