#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include "draw.h"
#include "cmd.h"
#include "gfx.h"
#include "area.h"
#include <unistd.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <signal.h>
#include "IMG_savepng.h"

#ifdef WIN32
#define THREADREAD
#endif

int yyparse(void);
int eventreporting = 0;
int yywrap() {
	return 1;
}

/* whether to keep processing events after EOF on stdin */
int keepalive = 1;
int quitonq = 0;
int scrwidth = 640, scrheight = 480, scrdepth = 0;
FILE *screendumpf = NULL;

int noclrscr = 0;
int headless = 0;

/* saved frame count */
int saveframecount = 0;
/* file name header for saved frames */
const char *saveframehead = NULL;

matrix_t eventmatrix = (matrix_t) { .v = { 1.0, 0, 0, 0, 1.0, 0, 0, 0, 1.0 } };

static void help()
{
	printf("Animator\n\n"
"-h        display this help\n"
"-x <N>    set horizontal resolution (default 640)\n"
"-y <N>    set vertical resolution (default 480)\n"
"-D <bpp>  force color depth (bits per pixel)\n"
"-k        disable keepalive, exit on EOF\n"
"-e        enable event reporting\n"
"-a        disable enforcing aspect ratio\n"
"-d <head> save all frames as headXXXX.png\n"
"-q        quit Animator when Q is pressed\n"
"-C        don't clear screen on new frame\n"
"-R        like -C, but allows resize\n"
"-H        headless mode (implies -k)\n"
"-F <N>    set screendump target fd\n"
);
}

void resize(int w, int h)
{
	SDL_Surface *win;

	if(headless) {
		win = SDL_CreateRGBSurface(SDL_SWSURFACE, w, h, 32, 0x00ff0000, 0x0000ff00, 0x000000ff, 0);
	} else if(noclrscr) {
		if(scrdepth == 0)
			win = SDL_SetVideoMode(w, h, 32, SDL_ANYFORMAT | SDL_SWSURFACE | (noclrscr == 2 ? SDL_RESIZABLE : 0));
		else
			win = SDL_SetVideoMode(w, h, scrdepth, SDL_SWSURFACE | (noclrscr == 2 ? SDL_RESIZABLE : 0));
	} else {
		if(scrdepth == 0)
			win = SDL_SetVideoMode(w, h, 32, SDL_RESIZABLE | SDL_ANYFORMAT | SDL_HWSURFACE | SDL_DOUBLEBUF);
		else
			win = SDL_SetVideoMode(w, h, scrdepth, SDL_RESIZABLE | SDL_HWSURFACE | SDL_DOUBLEBUF);
	}

	if(!win) {
		fprintf(stderr, "SDL_SetVideoMode failed\n");
		exit(1);
	}
	gfx_surface_sw(win);
	gfx_colorsupdate_serial ++;

	flushframe(0);
	areamap_resize(w / 2, h / 2);
}

static void printcoord(int x, int y)
{
	double ox, oy;
	untransform(x, y, &ox, &oy);
	printf(" %f %f", ox, oy);
}

static void printareas(int x, int y)
{
	area_list_t *a = areamap_query(x/2, y/2);
	if(!a) return;

	int i;
	for(i=0; i<a->count; i++) {
		printf(" %s", a->areas[i]->name);
	}
}

static void pump()
{
	SDL_Event e;

	SDL_PumpEvents();

	int nev=0;
	while(SDL_PeepEvents(&e, 1, SDL_GETEVENT, SDL_VIDEORESIZEMASK) == 1) nev ++;
	if(nev > 0) {
		/* event reporting has to come before resize() as resize() may trigger new events and order of reports would be reverse */
		if(eventreporting) {
			printf("RESIZE %d %d", e.resize.w, e.resize.h);
			printcoord(0, main_surface->h - 1);
			printcoord(main_surface->w - 1, 0);
			printf("\n");
		}
		resize(e.resize.w, e.resize.h);
	}

	nev=0;
	while(SDL_PeepEvents(&e, 1, SDL_GETEVENT, SDL_MOUSEMOTIONMASK) == 1) nev ++;
	if(nev > 0) {
		if(eventreporting) {
			printf("MOTION %d", e.button.button);
			printcoord(e.motion.x, e.motion.y);
			printareas(e.button.x, e.button.y);
			printf("\n");
		}
	}

	while(SDL_PeepEvents(&e, 1, SDL_GETEVENT, ~0) == 1) {
		switch(e.type) {
			case SDL_MOUSEBUTTONDOWN:
				if(eventreporting) {
					printf("PRESS %d", e.button.button);
					printcoord(e.button.x, e.button.y);
					printareas(e.button.x, e.button.y);
					printf("\n");
				}
				break;
			case SDL_MOUSEBUTTONUP:
				if(eventreporting) {
					printf("RELEASE %d", e.button.button);
					printcoord(e.button.x, e.button.y);
					printareas(e.button.x, e.button.y);
					printf("\n");
				}
				break;
			case SDL_KEYDOWN:
				if(eventreporting) {
					printf("KEYDOWN %d %d %s\n", e.key.keysym.mod, e.key.keysym.unicode, SDL_GetKeyName(e.key.keysym.sym));
				}
				if(quitonq && e.key.keysym.sym == SDLK_q) 
					exit(0);
				break;
			case SDL_KEYUP:
				if(eventreporting) {
					printf("KEYUP %d %s\n", e.key.keysym.mod, SDL_GetKeyName(e.key.keysym.sym));
				}
				break;
			case SDL_QUIT:
				exit(0);
				break;
		}
	}

	if(eventreporting) fflush(stdout);
}

void saveframe()
{
	char buf[256];
	sprintf(buf, "%s%04d.png", saveframehead, saveframecount ++);
	if(IMG_SavePNG(buf, main_surface, -1)) {
		fprintf(stderr, "Sorry, failed to save PNG %s\n", buf);
		exit(1);
	}
}

void saveframeas(const char *name)
{
	render();
	if(IMG_SavePNG(name, main_surface, -1)) {
		fprintf(stderr, "Sorry, failed to save PNG %s\n", name);
		exit(1);
	}
	SDL_Flip(main_surface);

	pump();
}

void dumpframe(const char *name)
{
	int y;
	unsigned char *ptr;

	render();

	FILE *fp;
	if(name == NULL) {
		fp = screendumpf;
	} else {
		fp = fopen(name, "wb");
		if(fp == NULL) {
			perror("Failed to write screen dump file");
			fprintf(stderr, "filename: %s\n", name);
			return;
		}
	}

	Uint32 datalen = main_surface->format->BytesPerPixel * main_surface->w;
	if(datalen == main_surface->pitch) {
		if(fwrite(main_surface->pixels, datalen * main_surface->h, 1, fp) != 1) {
			perror("Failed to write screen dump");
		}
	} else {
		ptr = main_surface->pixels;
		for(y=0; y<main_surface->h; y++, ptr += main_surface->pitch) {
			if(fwrite(ptr, datalen, 1, fp) != 1) {
				perror("Failed to write screen dump");
				break;
			}
		}
	}

	if(name == NULL) {
		if(fflush(screendumpf))
			perror("Failed to write screen dump");
	} else {
		if(fclose(fp))
			perror("Failed to close screen dump file");
	}

	pump();
}

void flushframe(int command)
{
	render();
	if(command && saveframehead) saveframe();
	SDL_Flip(main_surface);

	pump();
}

void flushdelay(double tm)
{
	static double lastsleep = 0.0;
	static double lastframe = 0.0;

	double now = dbltime();
	double dt = 0;

	if(lastframe != 0.0)
		dt = now - lastframe;

	lastframe = now;

	double wait = tm - dt + lastsleep;
	if(wait < 0) wait = 0;

	flushframe(1);

	SDL_Delay(wait * 1000.0);
	lastsleep = wait;

	pump();

	if(eventreporting) {
		printf("FLUSHDELAY %g\n", dt);
		fflush(stdout);
	}
}

#ifdef THREADREAD
static SDL_Thread *readthread = NULL;

#define READBUF 65536
static char readbuf[READBUF];
static int readbuf_at = 0;
static int readbuf_eof = 0;
static SDL_mutex *readmutex = NULL;

/* this is the entry point of the reader thread */
static int readerthread(void *data)
{
#define INBUF 1024
	char inbuf[INBUF];
	ssize_t toread;

	for(;;) {
		if(readbuf_eof) {
			SDL_Delay(1000);
			continue;
		}

		SDL_mutexP(readmutex);
		toread = READBUF - readbuf_at;
		SDL_mutexV(readmutex);

		if(toread < INBUF) {
			SDL_Delay(10);
			continue;
		}

		ssize_t r = read(0, inbuf, INBUF);

		SDL_mutexP(readmutex);
		if(r < 1) {
			readbuf_eof = 1;
		} else {
			memcpy(readbuf + readbuf_at, inbuf, r);
			readbuf_at += r;
		}
		SDL_mutexV(readmutex);
	}

	return 0; /* not reached */
}

/* start reader thread */
static void readthread_start()
{
	readmutex = SDL_CreateMutex();
	readthread = SDL_CreateThread(readerthread, NULL);
}

/* shut down reader thread */
static void readthread_stop()
{
	if(readthread != NULL) SDL_KillThread(readthread);
}

/* in this mode, we do ordinary blocking reading in a separate thread */
int doinput(char *buf, int size)
{
	int r;

	for(;;) {
		pump();

		r = 0;

		SDL_mutexP(readmutex);
		if(readbuf_at > 0) {
			r = readbuf_at > size ? size : readbuf_at;
			memcpy(buf, readbuf, r);

			if(r < readbuf_at)
				memmove(readbuf, readbuf + r, readbuf_at - r);

			readbuf_at -= r;
		} else if(readbuf_eof) {
			r = -1;
		}

		SDL_mutexV(readmutex);

		if(r < 0) return 0;
		if(r > 0) return r;

		SDL_Delay(10);
	}
}
#else
/* normal reading using select */
int doinput(char *buf, int size)
{
	int r;

	for(;;) {
		pump();
		fd_set rfds;
		struct timeval tv;
		FD_ZERO(&rfds);
		FD_SET(0, &rfds);
		tv.tv_sec = 0;
		tv.tv_usec = 10000;
sel:	r = select(1, &rfds, NULL, NULL, &tv);
		if(r == -1 && errno == EINTR) goto sel;

		if(r == -1) {
			perror("select");
			return 0;
		} else {
			if(r) {
				int rr = read(0, buf, size);
				return rr;
			}
		}
	}
}
#endif

int main(int argc, char *argv[])
{
	screendumpf = stdout;

	int opt;
	while((opt = getopt(argc, argv, "hx:y:kead:D:qCRHF:")) != -1)
	{
		switch(opt) {
		case 'h':
			help();
			exit(0);

		case 'q':
			quitonq = 1;
			break;

		case 'x':
			scrwidth = atoi(optarg);
			if(scrwidth < 1) {
				fprintf(stderr, "Invalid horizontal resolution\n");
				exit(1);
			}
			break;

		case 'y':
			scrheight = atoi(optarg);
			if(scrheight < 1) {
				fprintf(stderr, "Invalid vertical resolution\n");
				exit(1);
			}
			break;

		case 'D':
			scrdepth = atoi(optarg);
			if(scrdepth != 0 && scrdepth != 8 && scrdepth != 15 && scrdepth != 16 && scrdepth != 24 && scrdepth != 32) {
				fprintf(stderr, "Invalid color depth\n");
				exit(1);
			}
			break;

		case 'F':
			screendumpf = fdopen(atoi(optarg), "wb");
			if(screendumpf == NULL) {
				perror("Error opening specified screen dump fd");
				exit(1);
			}
			break;

		case 'k':
			keepalive = 0;
			break;

		case 'e':
			eventreporting = 1;
			break;

		case 'a':
			keepaspect = 0;
			break;

		case 'd':
			saveframehead = optarg;
			if(strlen(saveframehead) > 192) {
				fprintf(stderr, "Your saved frame name header is too long, sorry\n");
				exit(1);
			}
			break;

		case 'C':
			noclrscr = 1;
			break;

		case 'R':
			noclrscr = 2;
			break;

		case 'H':
			headless = 1;
			keepalive = 0;
			break;

		default:
			fprintf(stderr, "Invalid argument, see -h\n");
			exit(1);
		}
	}

	if(optind < argc) {
		fprintf(stderr, "Superfluous arguments:\n");
		while(optind < argc) {
			fprintf(stderr, "  %s\n", argv[optind++]);
		}
		exit(1);
	}

	if(headless) {
		/* make headless mode not need DISPLAY */
		putenv("SDL_VIDEODRIVER=dummy");
	}

	if(SDL_Init(SDL_INIT_VIDEO) == -1) {
		fprintf(stderr, "SDL_Init failed\n");
		exit(1);
	}

	if(!headless) {
		SDL_EnableUNICODE(1);
		SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);
	}

	atexit(SDL_Quit);

	init_area();
	gfx_init();
	init_macros();
	image_init();
	cursor_init();

	resize(scrwidth, scrheight);

	if(!headless)
		SDL_WM_SetCaption("Animator", "Animator");

#ifdef THREADREAD
	readthread_start();
#else
	if(fcntl(0, F_SETFD, fcntl(0, F_GETFD, 0) | O_NONBLOCK))
	{
		perror("fcntl");
		exit(1);
	}
#endif

	pushmatrix();

	signal(SIGTERM, SIG_DFL);

	yyparse();

	if(keepalive) {
		for(;;) {
			pump();
			SDL_Delay(15);
		}
	}

#ifdef THREADREAD
	readthread_stop();
#endif

	return 0;
}
