#include <stdlib.h>
#include "util_vptrhash.h"

#define vptrht_MINSIZE 8
#define vptrht_MAXSIZE (1U << 31)

#define JUMP(i, j) i += j++
#define JUMP_FIRST(i, j) j = 1, i += j++

static inline int keyeq(vptrht_key_t a, vptrht_key_t b) {
	return a == b;
}

static inline unsigned int keyhash(vptrht_key_t k) {
	unsigned long n = (unsigned long)k;
	return (n >> 2) ^ (n >> 12);

}

static inline unsigned int entryhash(const vptrht_entry_t *entry) {
	return keyhash(entry->key);
}

vptrht_t *vptrht_alloc(void) {
	vptrht_t *ht = malloc(sizeof(vptrht_t));

	vptrht_init(ht);
	return ht;
}

void vptrht_init(vptrht_t *ht) {
	ht->mask = vptrht_MINSIZE - 1;
	ht->fill = 0;
	ht->table = calloc(ht->mask + 1, sizeof(vptrht_entry_t));
}

void vptrht_clear(vptrht_t *ht) {
	vptrht_uninit(ht);
	vptrht_init(ht);
}

void vptrht_uninit(vptrht_t *ht) {
	free(ht->table);
	ht->table = NULL;
}

void vptrht_free(vptrht_t *ht) {
	vptrht_uninit(ht);
	free(ht);
}

/* one lookup function to rule them all */
static inline vptrht_entry_t *vptrht_lookup(vptrht_t *ht, vptrht_key_t key, vptrht_hash_t hash) {
	unsigned int mask = ht->mask;
	unsigned int i = hash;
	unsigned int j;
	vptrht_entry_t *table = ht->table;
	vptrht_entry_t *entry = table + (i & mask);

	/* two separate if's for easier profiling */
	if (vptrht_isempty(entry))
		return entry;
	if (keyeq(entry->key, key))
		return entry;
	for (JUMP_FIRST(i, j); ; JUMP(i, j)) {
		entry = table + (i & mask);
		if (vptrht_isempty(entry))
			return entry;
		if (keyeq(entry->key, key))
			return entry;
	}
}

/* for copy and resize: entry->key is not in ht */
static void cleaninsert(vptrht_t *ht, const vptrht_entry_t *entry) {
	unsigned int mask = ht->mask;
	unsigned int i = entryhash(entry);
	unsigned int j;
	vptrht_entry_t *table = ht->table;
	vptrht_entry_t *newentry = table + (i & mask);

	if (!vptrht_isempty(newentry))
		for (JUMP_FIRST(i, j); !vptrht_isempty(newentry); JUMP(i, j))
			newentry = table + (i & mask);
	++ht->fill;
	*newentry = *entry;
}

void vptrht_resize(vptrht_t *ht, unsigned int hint) {
	unsigned int newsize;
	unsigned int oldfill = ht->fill;
	vptrht_entry_t *oldtable = ht->table;
	vptrht_entry_t *entry;

	if (vptrht_MAXSIZE - (vptrht_MAXSIZE >> 3) < oldfill)
		abort();
	if (hint < oldfill << 1)
		hint = oldfill << 1;
	if (hint > vptrht_MAXSIZE || hint <= oldfill)
		hint = vptrht_MAXSIZE;
	for (newsize = vptrht_MINSIZE; newsize < hint; newsize <<= 1);
	ht->table = calloc(newsize, sizeof(vptrht_entry_t));
	ht->mask = newsize - 1;
	ht->fill = 0;
	for (entry = oldtable; oldfill > 0; ++entry)
		if (!vptrht_isempty(entry)) {
			--oldfill;
			cleaninsert(ht, entry);
		}
	free(oldtable);
}

int vptrht_has(vptrht_t *ht, vptrht_key_t key) {
	vptrht_entry_t *entry = vptrht_lookup(ht, key, keyhash(key));

	return !vptrht_isempty(entry);
}

vptrht_value_t vptrht_get(vptrht_t *ht, vptrht_key_t key) {
	vptrht_entry_t *entry = vptrht_lookup(ht, key, keyhash(key));

	return entry->value; /* 0 if entry is empty */
}

/* fill threshold = 3/4 */
static inline void checkfill(vptrht_t *ht) {
	if (ht->fill > ht->mask - (ht->mask >> 2))
		vptrht_resize(ht, ht->fill << (ht->fill > 1 << 16 ? 1 : 2));
}

static vptrht_entry_t *vptrht_insert(vptrht_t *ht, vptrht_key_t key, vptrht_value_t value) {
	vptrht_hash_t hash = keyhash(key);
	vptrht_entry_t *entry = vptrht_lookup(ht, key, hash);

	if (!vptrht_isempty(entry))
		return entry;
	++ht->fill;
	entry->key = key;
	entry->value = value;
	checkfill(ht);
	return NULL;
}

void vptrht_set(vptrht_t *ht, vptrht_key_t key, vptrht_value_t value) {
	vptrht_entry_t *entry = vptrht_insert(ht, key, value);

	if (entry)
		entry->value = value;
}
